@@ -6,6 +6,7 @@
 #define _ATMARPD_H
 
 #include <stdint.h>
+#include <atm.h>
 
 
 #define ATMARP_SOCKET_PATH	"/dev/atmarp"	/* it seems awfully silly to
@@ -20,6 +21,7 @@
 
 #define ATF_NULL	0x1000	/* use NULL encapsulation */
 #define ATF_ARPSRV	0x2000	/* entry describes ARP server */
+#define ATF_NOVC	0x4000	/* query only; do not create a VC */
 
 
 enum atmarp_req_type {
@@ -28,7 +30,8 @@ enum atmarp_req_type {
 	art_qos,		/* set the default QoS */
 	art_set,		/* create or change an entry */
 	art_delete,		/* delete an entry */
-	art_table		/* update the ATMARP table file */
+	art_table,		/* update the ATMARP table file */
+	art_query		/* request resolution without VC setup */
 };
 
 struct atmarp_req {
