@@ -35,12 +35,20 @@
 
 AsnOid atmfPortIndex = {13, "\53\06\01\04\01\202\141\02\01\01\01\01\00"};
 AsnOid atmfMyIpNmAddress = {11, "\53\06\01\04\01\202\141\02\01\02\00"};
+AsnOid atmfAtmLayerMaxVpiBits =  {13,        "\53\06\01\04\01\202\141\02\02\01\01\6\00"};
+AsnOid atmfAtmLayerMaxVciBits =  {13,        "\53\06\01\04\01\202\141\02\02\01\01\7\00"};
 AsnOid atmfAtmLayerUniVersion =  {13,        "\53\06\01\04\01\202\141\02\02\01\01\11\00"};
 AsnOid atmfNetPrefixStatus = {NETPREFIX_LEN, "\53\06\01\04\01\202\141\02\07\01\01\03"};
 
 AsnInt atmfPortIndexValue = 0;
 IpAddress atmfMyIpNmAddressValue = {4 , NULL};
 
+/* The following two values depend on the capabilities of both
+   the switch AND the adapter - DO THEY ?? */
+AsnInt atmfAtmLayerMaxVpiBitsValue;
+AsnInt atmfAtmLayerMaxVciBitsValue;
+
+
 #if defined(UNI30) || defined(DYNAMIC_UNI)
 AsnInt atmfAtmLayerUniVersionValue = 2;		// version3point0(2)
 #else
