@@ -57,10 +57,11 @@ void ilmi_loop(int fd, int itf);
 void usage(char *name)
 {
   fprintf(stderr, "usage: %s [-b] [-d] [-v] [-l logfile] [-x] [-q qos] "
+    "%13s[ -i local_ip ] "
 #ifdef DYNAMIC_UNI
     "[ -u uni ] "
 #endif
-    "<interface>\n", name);
+    "<interface>\n",name,"");
   exit(1);
 }
 
@@ -272,7 +273,11 @@ int main(int argc, char *argv[])
   set_application("ilmid");
   set_verbosity(NULL, DIAG_WARN);
   qos = NULL;
-  while((opt = getopt(argc, argv, "bdhl:q:xv")) != EOF)
+  while((opt = getopt(argc, argv, "bdhi:l:q:xv"
+#ifdef DYNAMIC_UNI
+    "u"
+#endif
+    )) != EOF)
     switch (opt)
       {
       case 'd':
@@ -284,6 +289,14 @@ int main(int argc, char *argv[])
       case 'b':
 	bg = 1;
 	break;
+      case 'i':
+	{
+	  uint32_t ip;
+
+	  ip = text2ip(optarg,NULL,T2I_ERROR);
+	  if (!ip) return 1;
+	  set_local_ip(ip);
+	}
       case 'l':
 	set_logfile(optarg);
 	break;
