@@ -37,6 +37,7 @@
 #include <stdlib.h>  /* for malloc() and free() */
 /* Atm includes */
 #include <atm.h>
+#include <linux/atmdev.h>
 #include <linux/atmlec.h>
 
 /* Local includes */
@@ -99,7 +100,8 @@ kernel_init(unsigned char *mac_addr, int itf_num)
 int 
 kernel_sendmsg(atmlec_msg_type type, unsigned char *mac_addr, 
 	       unsigned char *atm_addr, struct atmlec_config_msg *config,
-	       unsigned long flag, char *data, int datalen)
+	       unsigned long flag, char *data, int datalen,
+               int targetless_le_arp, int no_source_le_narp)
 {
   struct atmlec_msg *mesg;
 
@@ -113,6 +115,8 @@ kernel_sendmsg(atmlec_msg_type type, unsigned char *mac_addr,
     if (atm_addr)
       memcpy(&mesg->content.normal.atm_addr, atm_addr, ATM_ESA_LEN);
     mesg->content.normal.flag = flag;
+    mesg->content.normal.targetless_le_arp = targetless_le_arp; /* LANE2 */
+    mesg->content.normal.no_source_le_narp = no_source_le_narp;
   } else
     memcpy(&mesg->content.config, config, sizeof(struct atmlec_config_msg));
 
@@ -161,6 +165,8 @@ get_mesg_type_str(atmlec_msg_type type)
     return "ARP_UPDATE";
   case l_config:
     return "CONFIG";
+  case l_associate_req:
+    return "LANE2_ASSOCIATE_REQ";
   default:
     return "UNKNOWN MSG TYPE";
   }
@@ -193,9 +199,7 @@ int read_from_kernel (char *buff, int size)
 {
         int retval;
         
-        fprintf(stderr, "read_from_kernel: about to read\n");
         retval = read(lec_socket, buff, size);
-        fprintf(stderr, "read_from_kernel: read returned %d\n", retval);
         if (retval < 0)
                 EVENT(EM_SERR,("Read for extra data from kernel socket failed:%s\n",strerror(errno)));
         return retval;
