@@ -10,13 +10,25 @@
 #include "atm.h"
 
 
-int atm_equal(const struct sockaddr_atmsvc *a,const struct sockaddr_atmsvc *b,
-  int len,int flags)
+static int atm_equal_pvc(const struct sockaddr_atmpvc *a,
+  const struct sockaddr_atmpvc *b,int flags)
+{
+    int wc = flags & AXE_WILDCARD;
+
+#define EQ(field,any) \
+  (a->sap_addr.field == b->sap_addr.field || \
+    (wc && (a->sap_addr.field == any || b->sap_addr.field == any)))
+    return EQ(itf,ATM_ITF_ANY) && EQ(vpi,ATM_VPI_ANY) && EQ(vci,ATM_VCI_ANY);
+#undef EQ
+}
+
+
+static int atm_equal_svc(const struct sockaddr_atmsvc *a,
+  const struct sockaddr_atmsvc *b,int len,int flags)
 {
     const unsigned char *a_prv,*b_prv;
     int len_a,len_b;
 
-    assert(a->sas_family == AF_ATMSVC && b->sas_family == AF_ATMSVC);
     if (!(flags & AXE_WILDCARD)) len = ATM_ESA_LEN*8;
     assert(len >= 0 && len <= ATM_ESA_LEN*8);
     if (*a->sas_addr.prv && *b->sas_addr.prv) {
@@ -57,3 +69,16 @@ int atm_equal(const struct sockaddr_atmsvc *a,const struct sockaddr_atmsvc *b,
     return !strncmp(a->sas_addr.pub,b->sas_addr.pub,len_a < len_b ? len_a :
       len_b);
 }
+
+
+int atm_equal(const struct sockaddr *a,const struct sockaddr *b,int len,
+  int flags)
+{
+    assert((a->sa_family == AF_ATMPVC && b->sa_family == AF_ATMPVC) ||
+      (a->sa_family == AF_ATMSVC && b->sa_family == AF_ATMSVC));
+    if (a->sa_family == AF_ATMPVC)
+	return atm_equal_pvc((const struct sockaddr_atmpvc *) a,
+	  (const struct sockaddr_atmpvc *) b,flags);
+    return atm_equal_svc((const struct sockaddr_atmsvc *) a,
+      (const struct sockaddr_atmsvc *) b,len,flags);
+}
