@@ -31,68 +31,82 @@ extern FILE *yyin;
 
 int debug = 0;
 int pretty = A2T_PRETTY | A2T_NAME | A2T_LOCAL;
-int sig_pcr = -1; /* obsolete */
-const char *sig_qos = NULL;
 const char *dump_dir = NULL;
 
-SIGNALING_MODE mode = sm_user;
-int net = 0;
-int allocate_ci = 1;
+
+/* A little hack until we have full support for multiple signaling entities */
+
+SIG_ENTITY _entity = {
+	0,		/* fd */
+	0,		/* unspecified UNI version */
+	sm_user,	/* mode */
+	-1,		/* sig_pcr; obsolete @@@ */
+	NULL		/* sig_qos */
+};
 
 
 /* ------------------------------ SAAL relays ------------------------------ */
 
 
-static SAAL_DSC saal;
-
-
 static void q_estab_conf(void *user_data,void *uu_data,int uu_length)
 {
-    saal_okay();
+    SIG_ENTITY *sig = user_data;
+
+    saal_okay(sig);
 }
 
 
 static void q_rel_ind(void *user_data,void *uu_data,int uu_length)
 {
-    saal_failure();
-    saal_estab_req(&saal,NULL,0);
+    SIG_ENTITY *sig = user_data;
+
+    saal_failure(sig);
+    saal_estab_req(&sig->saal,NULL,0);
 }
 
 
 static void q_restart(void *user_data,void *uu_data,int uu_length,int ind)
 {
-    saal_failure();
-    if (!ind) saal_okay(); /* actually, ind should probably never be zero */
+    SIG_ENTITY *sig = user_data;
+
+    saal_failure(sig);
+    if (!ind) saal_okay(sig);
+	/* actually, ind should probably never be zero */
 }
 
 
-void from_net(void *buffer,int size)
+void from_net(SIG_ENTITY *sig,void *buffer,int size)
 {
-    saal_pdu(&saal,buffer,size);
+    saal_pdu(&sig->saal,buffer,size);
 }
 
 
-void to_signaling(void *msg,int size)
+void to_signaling(SIG_ENTITY *sig,void *msg,int size)
 {
-    trace_uni("TO NETWORK",msg,size);
-    diag(COMPONENT,DIAG_DEBUG,"TO SAAL: %s (0x%02x) CR 0x%06x (%d bytes)",
+    trace_uni("TO NETWORK",sig,msg,size);
+    diag(COMPONENT,DIAG_DEBUG,"TO SAAL (%d.%d.%d): %s (0x%02x) CR 0x%06x "
+      "(%d bytes)",S_PVC(sig),
       mid2name(((unsigned char *) msg)[5]),((unsigned char *) msg)[5],
       (((unsigned char *) msg)[2] << 16) | (((unsigned char *) msg)[3] << 8) |
       ((unsigned char *) msg)[4],size);
-    saal_send(&saal,msg,size);
+    saal_send(&sig->saal,msg,size);
 }
 
 
 static void q_data_ind(void *user_data,void *data,int length)
 {
-    trace_uni("FROM NETWORK",data,length);
-    to_uni(data,length);
+    SIG_ENTITY *sig = user_data;
+
+    trace_uni("FROM NETWORK",sig,data,length);
+    to_uni(sig,data,length);
 }
 
 
 static void q_cpcs_send(void *user_data,void *data,int length)
 {
-    to_net(data,length);
+    SIG_ENTITY *sig = user_data;
+
+    to_net(sig,data,length);
 }
 
 
@@ -151,16 +165,20 @@ static void dump_sap(FILE *file,const char *label,struct atm_sap *sap)
 
 static void dump_status(FILE *file,const char *banner)
 {
+    SIG_ENTITY *sig;
     SOCKET *walk;
 
-    if (sockets) fprintf(file,"%s\n\n",banner);
-    for (walk = sockets; walk; walk = walk->next) {
-	fprintf(file,"0x%lx: %s, CR 0x%06lX, PVC %d.%d.%d\n",walk->id,
-	  state_name[walk->state],walk->call_ref,walk->pvc.sap_addr.itf,
-	  walk->pvc.sap_addr.vpi,walk->pvc.sap_addr.vci);
-        dump_addr(file,"local ",&walk->local);
-        dump_addr(file,"remote",&walk->remote);
-        dump_sap(file,"sap",&walk->sap);
+    if (entities) fprintf(file,"%s\n\n",banner);
+    for (sig = entities; sig; sig = sig->next) {
+	fprintf(file,"--- Entity %d.%d.%d ---\n",S_PVC(sig));
+	for (walk = sockets; walk; walk = walk->next) {
+	    fprintf(file,"0x%lx: %s, CR 0x%06lX, PVC %d.%d.%d\n",walk->id,
+	      state_name[walk->state],walk->call_ref,walk->pvc.sap_addr.itf,
+	      walk->pvc.sap_addr.vpi,walk->pvc.sap_addr.vci);
+	    dump_addr(file,"local ",&walk->local);
+	    dump_addr(file,"remote",&walk->remote);
+	    dump_sap(file,"sap",&walk->sap);
+	}
     }
 }
 
@@ -280,6 +298,16 @@ static void trace_on_exit(int status,void *dummy)
 }
 
 
+static void manual_override(void)
+{
+     /*
+      * Gross hack to avoid changing the command-line parameters ... @@@
+      */
+     entities = &_entity;
+     _entity.next = NULL;
+}
+
+
 static void usage(const char *name)
 {
     fprintf(stderr,"usage: %s [ -b ] [ -c config_file ] [ -d ] "
@@ -292,16 +320,17 @@ static void usage(const char *name)
 
 int main(int argc,char **argv)
 {
+    SIG_ENTITY *sig;
     const char *config_file;
     char *end;
     int c,background;
+    int net = 0,allocate_ci = 1;
 
     set_application("atmsigd");
     config_file = CONFIG_FILE;
     dump_dir = NULL;
     background = 0;
-    memset(&signaling_pvc,0,sizeof(signaling_pvc));
-    signaling_pvc.sap_addr.vci = 5;
+    memset(&_entity.signaling_pvc,0,sizeof(_entity.signaling_pvc));
     /* 1st pass to get the -c option */
     while ((c = getopt(argc,argv,"Abc:dD:l:m:nNP:q:t:")) != EOF)
 	if (c == 'c') config_file = optarg;
@@ -309,11 +338,15 @@ int main(int argc,char **argv)
 	diag(COMPONENT,DIAG_WARN,"%s not found. - Using defaults.",config_file);
     else if (yyparse())
 	    diag(COMPONENT,DIAG_FATAL,"Error in config file. - Aborting.");
+    if (!atmpvc_addr_in_use(_entity.signaling_pvc))
+	_entity.signaling_pvc.sap_addr.vci = 5;
+    if (!_entity.uni) _entity.uni = S_UNI30;
     /* process all other options but -c */
     optind = 0;
     while ((c = getopt(argc,argv,"Abc:dD:l:m:nNP:q:t:")) != EOF)
 	switch (c) {
 	    case 'A':
+		manual_override();
 		allocate_ci = 0;
 		break;
 	    case 'b':
@@ -337,25 +370,31 @@ int main(int argc,char **argv)
 		set_logfile(optarg);
 		break;
 	    case 'm':
-		if (!strcmp(optarg,"user")) mode = sm_user;
-		else if (!strcmp(optarg,"network")) mode = sm_net;
-		else if (!strcmp(optarg,"switch")) mode = sm_switch;
+		manual_override();
+		if (!strcmp(optarg,"user")) _entity.mode = sm_user;
+		else if (!strcmp(optarg,"network")) _entity.mode = sm_net;
+		else if (!strcmp(optarg,"switch")) _entity.mode = sm_switch;
 		else usage(argv[0]);
 		break;
 	    case 'n':
 		pretty = A2T_PRETTY;
 		break;
 	    case 'N':
+		manual_override();
 		net = 1;
 		break;
 	    case 'q':
-		if (sig_pcr != -1) usage(argv[0]);
-		sig_qos = optarg;
+		manual_override();
+		if (_entity.sig_pcr != -1) usage(argv[0]);
+		_entity.sig_qos = optarg;
 		break;
 	    case 'P': /* obsolete */
-		if (sig_qos) usage(argv[0]);
-		sig_pcr = strtol(optarg,&end,0);
+		manual_override();
+		if (_entity.sig_qos) usage(argv[0]);
+		_entity.sig_pcr = strtol(optarg,&end,0);
 		if (*end) usage(argv[0]);
+		diag(COMPONENT,DIAG_WARN,"option -P is obsolete, "
+		  "please use  -q qos  instead");
 		break;
 	    case 't':
 		trace_size = strtol(optarg,&end,0);
@@ -364,41 +403,25 @@ int main(int argc,char **argv)
 	    default:
 		usage(argv[0]);
 	}
-    switch (mode) {
-	case sm_unknown:
-	    if (net) {
-		if (allocate_ci) {
-		    mode = sm_net;
-		    diag(COMPONENT,DIAG_WARN,"option -N is obsolete, "
-		      "please use  -m network  instead");
-		}
-		else {
-		    mode = sm_switch;
-		    diag(COMPONENT,DIAG_WARN,"options -N -A are obsolete, "
-		      "please use  -m switch  instead");
-		}
+    if (_entity.mode == sm_unknown)
+	if (net) {
+	    if (allocate_ci) {
+		_entity.mode = sm_net;
+		diag(COMPONENT,DIAG_WARN,"option -N is obsolete, "
+		  "please use  -m network  instead");
 	    }
-	    else if (allocate_ci) mode = sm_user;
-		else usage(argv[0]);
-	    break;
-	case sm_user:
-	    allocate_ci = 1; /* should be ignored anyway */
-	    net = 0;
-	    break;
-	case sm_net:
-	    allocate_ci = 1;
-	    net = 1;
-	    break;
-	case sm_switch:
-	    allocate_ci = 0;
-	    net = 1;
-	    break;
-	default:
-	    abort();
-    }
+	    else {
+		_entity.mode = sm_switch;
+		diag(COMPONENT,DIAG_WARN,"options -N -A are obsolete, "
+		  "please use  -m switch  instead");
+	    }
+	}
+	else if (allocate_ci) _entity.mode = sm_user;
+	    else usage(argv[0]);
     if (optind < argc) {
-	if (text2atm(argv[optind],(struct sockaddr *) &signaling_pvc,
-	  sizeof(signaling_pvc),T2A_PVC) < 0)
+	manual_override();
+	if (text2atm(argv[optind],(struct sockaddr *) &_entity.signaling_pvc,
+	  sizeof(_entity.signaling_pvc),T2A_PVC) < 0)
 	    diag(COMPONENT,DIAG_FATAL,"text2atm \"%s\": failed",argv[optind]);
 	optind++;
 	if (optind == argc-1) {
@@ -407,41 +430,33 @@ int main(int argc,char **argv)
 	}
     }
     if (optind != argc) usage(argv[0]);
-    diag(COMPONENT,DIAG_INFO,"Linux ATM signaling "
-#ifdef UNI30
-      "UNI 3.0"
-#endif
-#ifdef UNI31
-      "UNI 3.1"
-#ifdef ALLOW_UNI30
-      "+3.0compat"
-#endif
-#endif
-#ifdef UNI40
-      "UNI 4.0"
-#ifdef Q2963_1
-      "+Q.2931.1"
-#endif
-#endif
-      "/AAL5, version " VERSION);
     if (dump_dir)
 	if (chdir(dump_dir) < 0)
 	    diag(COMPONENT,DIAG_ERROR,"chdir %s: %s",dump_dir,strerror(errno));
-    diag(COMPONENT,DIAG_INFO,"Acting as %s",
-      mode == sm_user ? "USER side" :
-      mode == sm_net ? "NETWORK side" : "SWITCH");
+    for (sig = entities; sig; sig = sig->next) {
+	diag(COMPONENT,DIAG_INFO,"Linux ATM signaling %s"
+#ifdef DYNAMIC_UNI
+	  " (dynamic)"
+#endif
+	  ", version " VERSION " on %d.%d.%d",
+	  sig->uni & S_UNI30 ?
+	    sig->uni & S_UNI31 ? "UNI 3.1+3.0compat" : "UNI 3.0" :
+	    sig->uni & S_UNI31 ? "UNI 3.1" :
+	      sig->uni & S_Q2963_1 ? "UNI 4.0+Q.2963.1" : "UNI 4.0",
+          S_PVC(sig));
+	diag(COMPONENT,DIAG_INFO,"Acting as %s",
+	  sig->mode == sm_user ? "USER side" :
+	  sig->mode == sm_net ? "NETWORK side" : "SWITCH");
+    }
     if (open_all()) return 1;
     init_current_time();
-    if (mode != sm_switch) init_addr();
     q_start();
-    start_saal(&saal,&ops,NULL,
-#ifdef UNI30
-      sscop_qsaal1
-#else
-      sscop_q2110
-#endif
-      );
-    saal_estab_req(&saal,NULL,0);
+    for (sig = entities; sig; sig = sig->next) {
+	if (sig->mode != sm_switch) init_addr(sig);
+	start_saal(&sig->saal,&ops,sig,
+	  sig->uni == S_UNI30 ? sscop_qsaal1 : sscop_q2110);
+	saal_estab_req(&sig->saal,NULL,0);
+    }
     setup_signals();
     if (background) {
 	pid_t pid;
@@ -457,6 +472,6 @@ int main(int argc,char **argv)
     (void) on_exit(trace_on_exit,NULL);
     poll_loop();
     close_all();
-    stop_saal(&saal);
+    for (sig = entities; sig; sig = sig->next) stop_saal(&sig->saal);
     return 0;
 }
