@@ -43,35 +43,24 @@ static int send_setup(SOCKET *sock)
     while (walk);
     sock->call_ref = call_ref;
     q_create(&dsc,q_buffer,MAX_Q_MSG);
-    q_assign(&dsc,QF_call_ref,call_ref);
-    q_assign(&dsc,QF_msg_type,ATM_MSG_SETUP);
-    q_assign(&dsc,QF_aal_type,5); /* AAL 5 */
-#ifdef UNI30
-    q_assign(&dsc,QF_aal_mode,1); /* Message mode - LANE seems to really want
-				     this */
-#endif
-    q_assign(&dsc,QF_sscs_type,0); /* unspecified - LANE wants this */
-    error = sap_encode(&dsc,&sock->remote,&sock->sap,&sock->qos);
-    q_assign(&dsc,QF_bearer_class,16); /* BCOB-X */
-#if defined(UNI30) || defined(ALLOW_UNI30)
-    q_assign(&dsc,QF_trans_cap,ATM_TC_VBR_NRT_R00);
-	/* force presence - UNI 3.0 wants this */
-#endif
-    q_assign(&dsc,QF_upcc,0); /* p2p */
-#ifndef UNI30
-    q_assign(&dsc,QF_qos_cs,Q2931_CS_ITU);
-#endif
-    q_assign(&dsc,QF_qos_fw,0); /* QOS 0 */
-    q_assign(&dsc,QF_qos_bw,0);
     local = NULL;
+    error = 0;
     if (atmsvc_addr_in_use(sock->local)) local = &sock->local;
-    else if (local_addr[0].state == ls_same &&
-	  atmsvc_addr_in_use(local_addr[0].addr))
-	    local = &local_addr[0].addr;
-    if (!local) {
-	if (!error) error = -EADDRNOTAVAIL;
+    else {
+	sock->sig = route_call(&sock->remote);
+	if (!sock->sig) error = -EHOSTUNREACH;
+	else {
+	    local = get_local(sock->sig->signaling_pvc.sap_addr.itf);
+	    if (!local) local = get_local(ATM_ITF_ANY);
+	    if (local) sock->local = *local;
+	    else {
+		error = -EADDRNOTAVAIL;
+		diag(COMPONENT,DIAG_ERROR,"no local address");
+	    }
+	}
     }
-    else if (*local->sas_addr.pub) {
+    if (local)
+	if (*local->sas_addr.pub) {
 	    q_assign(&dsc,QF_cgpn_plan,ATM_NP_E164);
 	    q_assign(&dsc,QF_cgpn_type,ATM_TON_INTRNTNL);
 	    q_write(&dsc,QF_cgpn,(void *) local->sas_addr.pub,
@@ -82,26 +71,52 @@ static int send_setup(SOCKET *sock)
 	    q_assign(&dsc,QF_cgpn_type,ATM_TON_UNKNOWN);
 	    q_write(&dsc,QF_cgpn,(void *) local->sas_addr.prv,ATM_ESA_LEN);
 	}
-    if (net) {
+    q_assign(&dsc,QF_call_ref,call_ref);
+    q_assign(&dsc,QF_msg_type,ATM_MSG_SETUP);
+    q_assign(&dsc,QF_aal_type,5); /* AAL 5 */
+#if defined(UNI30) || defined(DYNAMIC_UNI)
+    if (sock->sig->uni == S_UNI30)
+	q_assign(&dsc,QF_aal_mode,1); /* Message mode - LANE seems to really
+					 want this */
+#endif
+    q_assign(&dsc,QF_sscs_type,0); /* unspecified - LANE wants this */
+    if (!error)
+	error = sap_encode(&dsc,&sock->remote,&sock->sap,&sock->qos,
+	  sock->sig->uni);
+    q_assign(&dsc,QF_bearer_class,16); /* BCOB-X */
+#if defined(UNI30) || defined(ALLOW_UNI30) || defined(DYNAMIC_UNI)
+    if (sock->sig->uni & S_UNI30)
+	q_assign(&dsc,QF_trans_cap,ATM_TC_VBR_NRT_R00);
+	  /* force presence - UNI 3.0 wants this */
+#endif
+    q_assign(&dsc,QF_upcc,0); /* p2p */
+#if !defined(UNI30) || defined(DYNAMIC_UNI)
+    if (sock->sig->uni != S_UNI30)
+	q_assign(&dsc,QF_qos_cs,Q2931_CS_ITU);
+#endif
+    q_assign(&dsc,QF_qos_fw,0); /* QOS 0 */
+    q_assign(&dsc,QF_qos_bw,0);
+    if (sock->sig->mode != sm_user) {
 	if (!atmpvc_addr_in_use(sock->pvc)) {
-	    int vci;
+	    int vpci,vci;
 
-	    if (!allocate_ci)
+	    if (sock->sig->mode == sm_switch)
 		diag(COMPONENT,DIAG_FATAL,"No CI allocator (use -A)");
-	    vci = get_vci(0);
+	    vpci = 0;
+	    sock->pvc.sap_addr.itf = get_itf(sock->sig,&vpci);
+	    sock->pvc.sap_addr.vpi = vpci;
+	    vci = get_vci(sock->pvc.sap_addr.itf);
 	    if (vci < 0) {
 		(void) q_close(&dsc);
 		return vci;
 	    }
-	    sock->pvc.sap_addr.itf = get_itf(0);
-	    sock->pvc.sap_addr.vpi = 0;
 	    sock->pvc.sap_addr.vci = vci;
 	}
 	q_assign(&dsc,QF_vpi,sock->pvc.sap_addr.vpi);
 	q_assign(&dsc,QF_vci,sock->pvc.sap_addr.vci);
     }
     if ((size = q_close(&dsc)) < 0) error = -EINVAL;
-    else if (!error) to_signaling(q_buffer,size);
+    else if (!error) to_signaling(sock->sig,q_buffer,size);
     return error;
 }
 
@@ -117,20 +132,21 @@ static int send_connect(SOCKET *sock)
     if (sock->ep_ref >= 0) q_assign(&dsc,QF_ep_ref,sock->ep_ref);
     if (sock->ep_ref <= 0) { /* no AAL parameters if ep_ref present and != 0 */
 	q_assign(&dsc,QF_aal_type,5);
-#ifdef UNI30
-	q_assign(&dsc,QF_aal_mode,1); /* Message mode - LANE seems to really
-					 want this */
+#if defined(UNI30) || defined(DYNAMIC_UNI)
+	if (sock->sig->uni == S_UNI30)
+	    q_assign(&dsc,QF_aal_mode,1); /* Message mode - LANE seems to
+					     really want this */
 #endif
 	q_assign(&dsc,QF_sscs_type,0); /* unspecified - LANE wants this */
 	q_assign(&dsc,QF_fw_max_sdu,sock->qos.rxtp.max_sdu);
 	q_assign(&dsc,QF_bw_max_sdu,sock->qos.txtp.max_sdu);
     }
-    if ((size = q_close(&dsc)) >= 0) to_signaling(q_buffer,size);
+    if ((size = q_close(&dsc)) >= 0) to_signaling(sock->sig,q_buffer,size);
     return 0;
 }
 
 
-#ifdef Q2963_1
+#if defined(Q2963_1) || defined(DYNAMIC_UNI)
 
 static void send_modify_request(SOCKET *sock)
 {
@@ -144,7 +160,7 @@ static void send_modify_request(SOCKET *sock)
 	q_assign(&dsc,QF_fw_pcr_01,SELECT_TOP_PCR(sock->new_qos.txtp));
     if (sock->new_qos.rxtp.traffic_class)
 	q_assign(&dsc,QF_bw_pcr_01,SELECT_TOP_PCR(sock->new_qos.rxtp));
-    if ((size = q_close(&dsc)) >= 0) to_signaling(q_buffer,size);
+    if ((size = q_close(&dsc)) >= 0) to_signaling(sock->sig,q_buffer,size);
 }
 
 
@@ -161,30 +177,12 @@ static void send_modify_ack(SOCKET *sock)
 	q_assign(&dsc,QF_type_of_report,ATM_TOR_MOD_CONF);
 	START_TIMER(sock,T361);
     }
-    if ((size = q_close(&dsc)) >= 0) to_signaling(q_buffer,size);
+    if ((size = q_close(&dsc)) >= 0) to_signaling(sock->sig,q_buffer,size);
 }
 
 #endif
 
 
-static int bind_check(SOCKET *sock,struct sockaddr_atmsvc *local)
-{
-    struct sockaddr_atmsvc *use;
-
-    if (atmsvc_addr_in_use(sock->local)) return 0;
-    if (atmsvc_addr_in_use(*local)) use = local;
-    else {
-	if (local_addr[0].state != ls_same) {
-	    diag(COMPONENT,DIAG_ERROR,"no local address");
-	    return -EADDRNOTAVAIL;
-	}
-	use = &local_addr[0].addr;
-    }
-    sock->local = *use;
-    return 0;
-}
-
-
 static void dispatch(SOCKET *sock,struct atmsvc_msg *msg)
 {
     int error;
@@ -215,9 +213,17 @@ static void dispatch(SOCKET *sock,struct atmsvc_msg *msg)
 		    if(*msg->svc.sas_addr.pub)
 			diag(COMPONENT,DIAG_DEBUG,"binding to E.164 address "
 			  "%s\n",msg->svc.sas_addr.pub);
-		if (walk->state == ls_unused)
+		if (walk->state == ls_unused) {
 		    SEND_ERROR(msg->vcc,-EADDRNOTAVAIL);
-		else send_kernel(msg->vcc,0,as_okay,0,NULL,NULL,NULL,NULL,NULL);
+		    return;
+		}
+		sock->sig = get_sig_entity(walk->itf);
+		if (!sock->sig) {
+		    diag(COMPONENT,DIAG_ERROR,"signaling entity lookup failed");
+		    SEND_ERROR(msg->vcc,-EADDRNOTAVAIL);
+		    return;
+		}
+		send_kernel(msg->vcc,0,as_okay,0,NULL,NULL,NULL,NULL,NULL);
 	    }
 	    return;
 	case as_connect: /* NULL state only */
@@ -227,12 +233,6 @@ static void dispatch(SOCKET *sock,struct atmsvc_msg *msg)
 		return;
 	    }
 	    sock = new_sock(msg->vcc);
-	    error = bind_check(sock,&msg->local);
-	    if (error) {
-		free_sock(sock);
-		SEND_ERROR(msg->vcc,error);
-		return;
-	    }
 	    sock->remote = msg->svc;
 	    sock->qos = msg->qos;
 	    sock->sap = msg->sap;
@@ -255,7 +255,7 @@ static void dispatch(SOCKET *sock,struct atmsvc_msg *msg)
 	    }
 	    if (sock->state != ss_indicated && sock->state != ss_proceeding)
 		break;
-	    if (sock->state == ss_indicated && net)
+	    if (sock->state == ss_indicated && sock->sig->mode != sm_user)
 		diag(COMPONENT,DIAG_FATAL,"No CI allocator (use -A)");
 	    error = send_connect(sock);
 	    if (!error) {
@@ -271,7 +271,8 @@ static void dispatch(SOCKET *sock,struct atmsvc_msg *msg)
 	case as_reject: /* ZOMBIE, INDICATED, or PROCEEDING */
 	    switch (sock->state) {
 		case ss_indicated:
-		    send_release_complete(sock->call_ref,ATM_CV_CALL_REJ);
+		    send_release_complete(sock->sig,sock->call_ref,
+		      ATM_CV_CALL_REJ);
 		    /* fall through */
 		case ss_zombie:
 		    free_sock(sock);
@@ -328,12 +329,12 @@ static void dispatch(SOCKET *sock,struct atmsvc_msg *msg)
 		    return;
 		case ss_connecting:
 		case ss_accepting:
-#ifdef Q2963_1
+#if defined(Q2963_1) || defined(DYNAMIC_UNI)
 		case ss_mod_req:
 #endif
 		    STOP_TIMER(sock);
 		    /* fall through */
-#ifdef Q2963_1
+#if defined(Q2963_1) || defined(DYNAMIC_UNI)
 		case ss_mod_lcl:
 		case ss_mod_rcv:
 		case ss_mod_fin_ok:
@@ -341,7 +342,7 @@ static void dispatch(SOCKET *sock,struct atmsvc_msg *msg)
 		case ss_mod_fin_ack:
 #endif
 		case ss_connected:
-#ifdef Q2963_1
+#if defined(Q2963_1) || defined(DYNAMIC_UNI)
 		    if (timer_handler(sock->conn_timer) == on_T361)
 			STOP_TIMER(sock);
 #endif
@@ -349,17 +350,22 @@ static void dispatch(SOCKET *sock,struct atmsvc_msg *msg)
 			diag(COMPONENT,DIAG_INFO,"Active close (CR 0x%06X)",
 			  sock->call_ref);
 		    send_release(sock,
+#ifdef DYNAMIC_UNI
+		      sock->sig->uni != S_UNI30 ? ATM_CV_NORMAL_CLEAR :
+		      ATM_CV_NORMAL_UNSPEC
+#else
 #if defined(UNI31) || defined(UNI40)
 		      ATM_CV_NORMAL_CLEAR
 #else
 		      ATM_CV_NORMAL_UNSPEC
+#endif
 #endif
 		      );
 		    START_TIMER(sock,T308_1);
 		    new_state(sock,ss_rel_req);
 		    return;
 		case ss_rel_ind:
-		    send_release_complete(sock->call_ref,0); /* @@@ */
+		    send_release_complete(sock->sig,sock->call_ref,0); /* @@@ */
 		    free_sock(sock);
 		    return;
 		default:
@@ -370,19 +376,24 @@ static void dispatch(SOCKET *sock,struct atmsvc_msg *msg)
 	    if (sock->state != ss_indicated && sock->state != ss_proceeding)
 		break;
 	    if (!atmpvc_addr_in_use(msg->pvc)) {
-		if (net && !allocate_ci)
+		if (sock->sig->mode == sm_switch)
 		    diag(COMPONENT,DIAG_FATAL,"No CI allocator (use -A)");
 		return;
 	    }
-	    if (net && allocate_ci)
+	    if (sock->sig->mode == sm_net)
 		diag(COMPONENT,DIAG_FATAL,"CI allocation role conflict");
 	    sock->pvc = msg->pvc;
 	    if (send_call_proceeding(sock))
 		diag(COMPONENT,DIAG_FATAL,"s_c_p failed");
 	    new_state(sock,ss_proceeding);
 	    return;
-#ifdef Q2963_1
+#if defined(Q2963_1) || defined(DYNAMIC_UNI)
 	case as_modify:
+	    if (sock && !(sock->sig->uni & S_Q2963_1)) {
+		send_kernel(sock->id,0,as_okay,-ENOPROTOOPT,NULL,NULL,NULL,
+		  NULL,NULL);
+		return;
+	    }
 	    if (sock && (sock->state == ss_mod_lcl ||
 	      sock->state == ss_mod_req || sock->state == ss_mod_rcv ||
 	      sock->state == ss_mod_fin_ok || sock->state == ss_mod_fin_fail ||
@@ -522,6 +533,10 @@ void itf_load(int itf)
 		    diag(COMPONENT,DIAG_INFO,"Removed local ATM address %s at "
 		      "itf %d",buf,itf);
 		    walk->state = ls_unused;
+/*
+ * This is probably wrong. What if we delete an entry from the middle of
+ * the list ? @@@
+ */
 		    /* @@@ delete SVCs using that address ? */
 		    break;
 		default:
@@ -541,14 +556,19 @@ void from_kernel(struct atmsvc_msg *msg,int size)
 	return;
     }
     if (msg->type == as_terminate) {
+#if 0 /* need to pass some ID ... @@@ */
 	if (mode != sm_switch) {
 	    diag(COMPONENT,DIAG_ERROR,"Ignoring as_terminate received in %s "
 	      "mode",mode == sm_user ? "USER" : "NETWORK");
 	    return;
 	}
-	clear_all_calls();
+	clear_all_calls(sig);
+#endif
 	stop = 1;
 	diag(COMPONENT,DIAG_INFO,"Going down on as_terminate");
+	/*
+	 * Fix this - we need to shut it down more gracefully.
+	 */
 	return;
     }
     if (msg->listen_vcc && (msg->type == as_accept || msg->type == as_reject ||
