@@ -39,7 +39,7 @@ const char *cs_name[] = {
 
 const char *as_name[] = { "<invalid>","as_bind","as_connect","as_accept",
   "as_reject","as_listen","as_okay","as_error","as_indicate","as_close",
-  "as_itf_notify","as_modify","as_identify" };
+  "as_itf_notify","as_modify","as_identify","as_terminate" };
 
 const CALL_STATE state_map[] = { /* formatting aligned with STATE */
 	cs_null,	cs_null,	cs_null,	cs_call_init,
@@ -58,6 +58,8 @@ const PARTY_STATE eps_map[] = {
 	ps_active,	ps_add_recv,	ps_active,	ps_active,	/* 8 */
 	ps_active };							/*12 */
 
+
+SIG_ENTITY *entities = &_entity;
 SOCKET *sockets = NULL;
 unsigned char q_buffer[MAX_Q_MSG];
 
@@ -67,8 +69,9 @@ SOCKET *new_sock(unsigned long id)
     SOCKET *sock;
  
     sock = alloc_t(SOCKET);
+    sock->sig = NULL;
     sock->state = ss_invalid;
-    sock->pvc.sap_addr.vpi = sock->pvc.sap_addr.vci = 0;
+    memset(&sock->pvc,0,sizeof(sock->pvc));
     sock->qos.txtp.traffic_class = sock->qos.rxtp.traffic_class = ATM_UBR;
     sock->id = id;
     memset(&sock->local,0,sizeof(sock->local));
@@ -250,20 +253,37 @@ void send_release(SOCKET *sock,unsigned char reason,...)
 	default:
     }
     va_end(ap);
-    if ((size = q_close(&dsc)) >= 0) to_signaling(q_buffer,size);
+    if ((size = q_close(&dsc)) >= 0) to_signaling(sock->sig,q_buffer,size);
 }
 
 
-void send_release_complete(unsigned long call_ref,unsigned char cause)
+void send_release_complete(SIG_ENTITY *sig,unsigned long call_ref,
+  unsigned char cause,...)
 {
+    va_list ap;
     Q_DSC dsc;
     int size;
  
     q_create(&dsc,q_buffer,MAX_Q_MSG);
     q_assign(&dsc,QF_msg_type,ATM_MSG_REL_COMP);
     q_assign(&dsc,QF_call_ref,call_ref);
-    if (cause) q_assign(&dsc,QF_cause,cause); /* @@@ more data */
-    if ((size = q_close(&dsc)) >= 0) to_signaling(q_buffer,size);
+    if (cause) {
+	q_assign(&dsc,QF_cause,cause);
+	va_start(ap,cause);
+	switch (cause) {
+	    case ATM_CV_INVALID_IE:
+		{
+		    unsigned char ie;
+
+		    ie = va_arg(ap,int);
+		    q_write(&dsc,QF_ie_id6,&ie,1);
+		}
+		break;
+	    default:
+	}
+	va_end(ap);
+    }
+    if ((size = q_close(&dsc)) >= 0) to_signaling(sig,q_buffer,size);
 }
 
 
@@ -276,7 +296,7 @@ void send_modify_reject(SOCKET *sock,unsigned char reason)
     q_assign(&dsc,QF_msg_type,ATM_MSG_MODIFY_REJ);
     q_assign(&dsc,QF_call_ref,sock->call_ref);
     q_assign(&dsc,QF_cause,reason);
-    if ((size = q_close(&dsc)) >= 0) to_signaling(q_buffer,size);
+    if ((size = q_close(&dsc)) >= 0) to_signaling(sock->sig,q_buffer,size);
 }
 
 
@@ -312,7 +332,7 @@ int get_vci(int itf)
     s = get_pvc(itf,&vci);
     if (s < 0) return s;
     for (walk = sockets; walk; walk = walk->next)
-	if (walk->pvc.sap_addr.vci == vci) {
+	if (walk->pvc.sap_addr.itf == itf && walk->pvc.sap_addr.vci == vci) {
 	    vci = get_vci(itf); /* this recursion will keep all the busy ones
 				   open until we return */
 	    break;
@@ -322,21 +342,11 @@ int get_vci(int itf)
 }
 
 
-typedef struct _vpci {
-    int vpci;
-    int itf;
-    struct _vpci *next;
-} VPCI;
-
-
-static VPCI *vpcis = NULL;
-
-
-void enter_vpci(int vpci,int itf)
+void enter_vpci(SIG_ENTITY *sig,int vpci,int itf)
 {
     VPCI *entry;
 
-    for (entry = vpcis; entry; entry = entry->next)
+    for (entry = sig->vpcis; entry; entry = entry->next)
 	if (entry->vpci == vpci) {
 	    diag(COMPONENT,DIAG_ERROR,"ignoring duplicate VPCI %d (itf %d)",
 	      vpci,itf);
@@ -345,29 +355,99 @@ void enter_vpci(int vpci,int itf)
     entry = alloc_t(VPCI);
     entry->vpci = vpci;
     entry->itf = itf;
-    entry->next = vpcis;
-    vpcis = entry;
+    entry->next = sig->vpcis;
+    sig->vpcis = entry;
 }
 
 
-int get_itf(int *vpci)
+int get_itf(SIG_ENTITY *sig,int *vpci)
 {
     VPCI *best,*walk;
 
     best = NULL;
-    for (walk = vpcis; walk; walk = walk->next)
+    for (walk = sig->vpcis; walk; walk = walk->next)
 	if (walk->vpci <= *vpci && (!best || best->vpci < walk->vpci))
 	    best = walk;
-    if (!best) return signaling_pvc.sap_addr.itf;
+    if (!best) return sig->signaling_pvc.sap_addr.itf;
     *vpci -= best->vpci;
     return best->itf;
 }
 
 
-void init_addr(void)
+void init_addr(SIG_ENTITY *sig)
 {
     VPCI *walk;
 
-    itf_load(signaling_pvc.sap_addr.itf);
-    for (walk = vpcis; walk; walk = walk->next) itf_load(walk->itf);
+    itf_load(sig->signaling_pvc.sap_addr.itf);
+    for (walk = sig->vpcis; walk; walk = walk->next) itf_load(walk->itf);
+}
+
+
+/*
+ * The following code is stolen from switch/route.c. Eventually, all this
+ * should move into a library.
+ */
+
+
+#include <limits.h>
+
+
+typedef struct _route {
+    struct sockaddr_atmsvc addr;
+    int len;
+    SIG_ENTITY *sig;
+    struct _route *next;
+} ROUTE;
+
+
+static ROUTE *routes = NULL;
+
+
+void add_route(SIG_ENTITY *sig,struct sockaddr_atmsvc *addr,int len)
+{
+    ROUTE *route;
+
+    for (route = routes; route; route = route->next)
+        if (route->len == len && (!len || atm_equal((struct sockaddr *) addr,
+          (struct sockaddr *) &route->addr,len,
+          AXE_PRVOPT | (len == INT_MAX ? 0 : AXE_WILDCARD))))
+            diag(COMPONENT,DIAG_FATAL,"duplicate route");
+    route = alloc_t(ROUTE);
+    if (addr) route->addr = *addr;
+    route->len = len;
+    route->sig = sig;
+    route->next = routes;
+    routes = route;
+}
+
+
+SIG_ENTITY *route_call(struct sockaddr_atmsvc *addr)
+{
+    ROUTE *best,*route;
+    int best_len;
+
+    if (!routes) return entities;
+    best = NULL;
+    best_len = -1;
+    for (route = routes; route; route = route->next)
+        if (route->len > best_len && (!route->len ||
+          atm_equal((struct sockaddr *) addr,(struct sockaddr *) &route->addr,
+          route->len,
+          AXE_PRVOPT | (route->len == INT_MAX ? 0 : AXE_WILDCARD)))) {
+            if (route->len == INT_MAX) return route->sig;
+            best_len = route->len;
+            best = route;
+        }
+    return best ? best->sig : NULL;
+}
+
+
+SIG_ENTITY *get_sig_entity(int itf)
+{
+    SIG_ENTITY *sig;
+
+    for (sig = entities; sig; sig = sig->next)
+	if (sig->signaling_pvc.sap_addr.itf == itf) return sig;
+    /* should also check for vpcis @@@ */
+    return NULL;
 }
