@@ -43,56 +43,63 @@ int send_call_proceeding(SOCKET *sock)
     q_create(&dsc,q_buffer,MAX_Q_MSG);
     q_assign(&dsc,QF_msg_type,ATM_MSG_CALL_PROC);
     q_assign(&dsc,QF_call_ref,sock->call_ref);
-    if (net && allocate_ci) {
-	int vci;
+    if (sock->sig->mode == sm_net) {
+	int vpci,vci;
 
-	vci = get_vci(0);
+	sock->pvc.sap_family = AF_ATMPVC;
+	vpci = sock->sig->signaling_pvc.sap_addr.itf;
+	sock->pvc.sap_addr.itf = get_itf(sock->sig,&vpci);
+	sock->pvc.sap_addr.vpi = vpci;
+	vci = get_vci(sock->pvc.sap_addr.itf);
 	if (vci < 0) {
 	    (void) q_close(&dsc);
 	    return vci;
 	}
-	sock->pvc.sap_family = AF_ATMPVC;
-	sock->pvc.sap_addr.itf = signaling_pvc.sap_addr.itf;
-	sock->pvc.sap_addr.vpi = 0;
 	sock->pvc.sap_addr.vci = vci;
     }
-    if (net) {
+    if (sock->sig->mode != sm_user) {
 	q_assign(&dsc,QF_vpi,sock->pvc.sap_addr.vpi);
 	q_assign(&dsc,QF_vci,sock->pvc.sap_addr.vci);
     }
     if (sock->ep_ref >= 0) q_assign(&dsc,QF_ep_ref,sock->ep_ref);
-    if ((size = q_close(&dsc)) >= 0) to_signaling(q_buffer,size);
+    if ((size = q_close(&dsc)) >= 0) to_signaling(sock->sig,q_buffer,size);
     return 0;
 }
 
 
-static void setup_call(unsigned long call_ref)
+static void setup_call(SIG_ENTITY *sig,unsigned long call_ref)
 {
     SOCKET *sock,*this,**walk;
     struct sockaddr_atmsvc in_addr;
     struct atm_sap in_sap;
     struct atm_qos in_qos;
+    unsigned int problem;
     int i;
 
-    sap_decode(&in_dsc,&in_addr,&in_sap,&in_qos);
+    problem = sap_decode(&in_dsc,&in_addr,&in_sap,&in_qos,sig->uni);
+    if (problem) {
+	send_release_complete(sig,call_ref,IE_PB_CAUSE(problem),
+	  IE_PB_IE(problem));
+	return;
+    }
     if (!atmsvc_addr_in_use(in_addr)) {
-	send_release_complete(call_ref,ATM_CV_UNALLOC);
+	send_release_complete(sig,call_ref,ATM_CV_UNALLOC);
 	return;
     }
     if (!allow(&in_addr,ACL_IN)) {
-	send_release_complete(call_ref,ATM_CV_REJ_CLIR);
+	send_release_complete(sig,call_ref,ATM_CV_REJ_CLIR);
 	return;
     }
     this = new_sock(0);
+    this->sig = sig;
     sock = lookup_sap(&in_addr,&in_sap,&in_qos,&this->local,&this->sap,
       &this->qos,0);
     if (!sock) {
 	free_sock(this);
-	send_release_complete(call_ref,ATM_CV_INCOMP_DEST);
+	send_release_complete(sig,call_ref,ATM_CV_INCOMP_DEST);
 	return;
     }
-    this->qos.aal = ATM_AAL5; /* hack @@@ */
-    this->state = net && allocate_ci ? ss_proceeding : ss_indicated;
+    this->state = sig->mode == sm_net ? ss_proceeding : ss_indicated;
     this->call_state = cs_in_proc;
     this->call_ref = call_ref;
     if (q_present(&in_dsc,QF_ep_ref))
@@ -100,24 +107,24 @@ static void setup_call(unsigned long call_ref)
 #ifdef CISCO
     else
 #endif
-    if (net && allocate_ci) {
+    if (sig->mode == sm_net) {
 	int error;
 
 	error = send_call_proceeding(this);
 	if (error) {
 	    free_sock(this);
-	    send_release_complete(call_ref,ATM_CV_NO_CI);
+	    send_release_complete(sig,call_ref,ATM_CV_NO_CI);
 	    return;
 	}
     }
     /* if (sock->local) *this->local->sas_addr = sock->local->sas_addr; ??? */
     diag(COMPONENT,DIAG_DEBUG,"AAL type %ld",q_fetch(&in_dsc,QF_aal_type));
-    if (!net) { /* already set by send_call_proceeding */
+    if (sig->mode == sm_user) { /* already set by send_call_proceeding */
 	int vpci;
 
 	vpci = q_fetch(&in_dsc,QF_vpi);
 	this->pvc.sap_family = AF_ATMPVC;
-	this->pvc.sap_addr.itf = get_itf(&vpci);
+	this->pvc.sap_addr.itf = get_itf(sig,&vpci);
 	this->pvc.sap_addr.vpi = vpci;
 	this->pvc.sap_addr.vci = q_fetch(&in_dsc,QF_vci);
     }
@@ -159,7 +166,7 @@ static void setup_call(unsigned long call_ref)
 }
 
 
-static void send_status(SOCKET *sock,unsigned long call_ref,
+static void send_status(SIG_ENTITY *sig,SOCKET *sock,unsigned long call_ref,
   unsigned char cause,...)
 {
     va_list ap;
@@ -200,7 +207,7 @@ static void send_status(SOCKET *sock,unsigned long call_ref,
 	    ;
     }
     va_end(ap);
-    if ((size = q_close(&dsc)) >= 0) to_signaling(q_buffer,size);
+    if ((size = q_close(&dsc)) >= 0) to_signaling(sig,q_buffer,size);
 }
 
 
@@ -213,7 +220,7 @@ static void send_status_enq(SOCKET *sock)
     q_assign(&dsc,QF_msg_type,ATM_MSG_STATUS_ENQ);
     q_assign(&dsc,QF_call_ref,sock->call_ref);
     if (sock->ep_ref >= 0) q_assign(&dsc,QF_ep_ref,sock->ep_ref);
-    if ((size = q_close(&dsc)) >= 0) to_signaling(q_buffer,size);
+    if ((size = q_close(&dsc)) >= 0) to_signaling(sock->sig,q_buffer,size);
     /* @@@ should start T322 */
 }
 
@@ -226,11 +233,11 @@ static void send_connect_ack(SOCKET *sock)
     q_create(&dsc,q_buffer,MAX_Q_MSG);
     q_assign(&dsc,QF_msg_type,ATM_MSG_CONN_ACK);
     q_assign(&dsc,QF_call_ref,sock->call_ref);
-    if ((size = q_close(&dsc)) >= 0) to_signaling(q_buffer,size);
+    if ((size = q_close(&dsc)) >= 0) to_signaling(sock->sig,q_buffer,size);
 }
 
 
-static void send_restart_ack(int vpi,int vci)
+static void send_restart_ack(SIG_ENTITY *sig,int vpi,int vci)
 {
     Q_DSC dsc;
     int size;
@@ -244,12 +251,12 @@ static void send_restart_ack(int vpi,int vci)
 	q_assign(&dsc,QF_vpi,vpi);
 	q_assign(&dsc,QF_vci,vci);
     }
-    if ((size = q_close(&dsc)) >= 0) to_signaling(q_buffer,size);
+    if ((size = q_close(&dsc)) >= 0) to_signaling(sig,q_buffer,size);
 }
 
 
-static void send_drop_party_ack(unsigned long call_ref,short ep_ref,
-  unsigned char cause)
+static void send_drop_party_ack(SIG_ENTITY *sig,unsigned long call_ref,
+  short ep_ref,unsigned char cause)
 {
     Q_DSC dsc;
     int size;
@@ -259,11 +266,11 @@ static void send_drop_party_ack(unsigned long call_ref,short ep_ref,
     q_assign(&dsc,QF_call_ref,call_ref);
     q_assign(&dsc,QF_ep_ref,ep_ref);
     q_assign(&dsc,QF_cause,cause);
-    if ((size = q_close(&dsc)) >= 0) to_signaling(q_buffer,size);
+    if ((size = q_close(&dsc)) >= 0) to_signaling(sig,q_buffer,size);
 }
 
 
-#ifdef Q2963_1
+#if defined(Q2963_1) || defined(DYNAMIC_UNI)
 
 static void send_conn_avail(SOCKET *sock)
 {
@@ -273,7 +280,7 @@ static void send_conn_avail(SOCKET *sock)
     q_create(&dsc,q_buffer,MAX_Q_MSG);
     q_assign(&dsc,QF_msg_type,ATM_MSG_CONN_AVAIL);
     q_assign(&dsc,QF_call_ref,sock->call_ref);
-    if ((size = q_close(&dsc)) >= 0) to_signaling(q_buffer,size);
+    if ((size = q_close(&dsc)) >= 0) to_signaling(sock->sig,q_buffer,size);
 }
 
 #endif
@@ -310,7 +317,8 @@ static void uni_call(SOCKET *sock,unsigned char mid)
     switch (mid) {
 	case ATM_MSG_CALL_PROC: /* CONNECTING, WAIT_REL, REL_REQ */
 	    if (sock->state == ss_wait_rel || sock->state == ss_rel_req) {
-		send_status(sock,0,ATM_CV_INCOMP_MSG,ATM_MSG_CALL_PROC);
+		send_status(sock->sig,sock,0,ATM_CV_INCOMP_MSG,
+		  ATM_MSG_CALL_PROC);
 		return;
 	    }
 	    if (sock->state != ss_connecting) break;
@@ -321,7 +329,7 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 
 		vpci = q_fetch(&in_dsc,QF_vpi);
 		sock->pvc.sap_family = AF_ATMPVC;
-		sock->pvc.sap_addr.itf = get_itf(&vpci);
+		sock->pvc.sap_addr.itf = get_itf(sock->sig,&vpci);
 		sock->pvc.sap_addr.vpi = vpci;
 		sock->pvc.sap_addr.vci = q_fetch(&in_dsc,QF_vci);
 		diag(COMPONENT,DIAG_DEBUG,"ITF.VPI.VCI: %d.%d.%d",
@@ -333,7 +341,7 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 	    return;
 	case ATM_MSG_CONNECT: /* CONNECTING, REL_REQ */
 	    if (sock->state == ss_rel_req) {
-		send_status(sock,0,ATM_CV_INCOMP_MSG,ATM_MSG_CONNECT);
+		send_status(sock->sig,sock,0,ATM_CV_INCOMP_MSG,ATM_MSG_CONNECT);
 		return;
 	    }
 	    if (sock->state != ss_connecting) break;
@@ -343,7 +351,7 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 
 		vpci = q_fetch(&in_dsc,QF_vpi);
 		sock->pvc.sap_family = AF_ATMPVC;
-		sock->pvc.sap_addr.itf = get_itf(&vpci);
+		sock->pvc.sap_addr.itf = get_itf(sock->sig,&vpci);
 		sock->pvc.sap_addr.vpi = vpci;
 		sock->pvc.sap_addr.vci = q_fetch(&in_dsc,QF_vci);
 		diag(COMPONENT,DIAG_DEBUG,"ITF.VPI.VCI: %d/%d.%d",
@@ -367,19 +375,20 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 	    send_kernel(sock->id,0,as_okay,0,&sock->pvc,NULL,&sock->local,
 	      &sock->sap,&sock->qos);
 	    new_state(sock,ss_connected);
-#ifdef Q2963_1
+#if defined(Q2963_1) || defined(DYNAMIC_UNI)
 	    sock->owner = 1;
 #endif
 	    if (atm2text(buffer,MAX_ATM_ADDR_LEN+1,(struct sockaddr *)
-	      &sock->local,0) < 0) strcpy(buffer,"<invalid>");
+	      &sock->remote,0) < 0) strcpy(buffer,"<invalid>");
 	    diag(COMPONENT,DIAG_INFO,"Active open succeeded (CR 0x%06X, "
-	      "ID 0x%08x, addr %s)",sock->call_ref,sock->id,buffer);
+	      "ID 0x%08x, to %s)",sock->call_ref,sock->id,buffer);
 	    return;
 	case ATM_MSG_CONN_ACK: /* ACCEPTING, WAIT_REL, REL_REQ */
 	    diag(COMPONENT,DIAG_DEBUG,"CA vpi.vci=%d.%d",
 	      sock->pvc.sap_addr.vpi,sock->pvc.sap_addr.vci);
 	    if (sock->state == ss_wait_rel || sock->state == ss_rel_req) {
-		send_status(sock,0,ATM_CV_INCOMP_MSG,ATM_MSG_CONN_ACK);
+		send_status(sock->sig,sock,0,ATM_CV_INCOMP_MSG,
+		  ATM_MSG_CONN_ACK);
 		return;
 	    }
 	    if (sock->state != ss_accepting) break;
@@ -387,13 +396,13 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 	    send_kernel(sock->id,0,as_okay,0,NULL,NULL,&sock->local,&sock->sap,
 	      NULL);
 	    new_state(sock,ss_connected);
-#ifdef Q2963_1
+#if defined(Q2963_1) || defined(DYNAMIC_UNI)
 	    sock->owner = 0;
 #endif
 	    if (atm2text(buffer,MAX_ATM_ADDR_LEN+1, (struct sockaddr *)
-	      &sock->local,0) < 0) strcpy(buffer,"<invalid>");
+	      &sock->remote,0) < 0) strcpy(buffer,"<invalid>");
 	    diag(COMPONENT,DIAG_INFO,"Passive open succeeded (CR 0x%06X, "
-	      "ID 0x%08x, addr %s)",sock->call_ref,sock->id,buffer);
+	      "ID 0x%08x, from %s)",sock->call_ref,sock->id,buffer);
 	    return;
 	case ATM_MSG_RELEASE: /* all states */
 	    {
@@ -409,7 +418,7 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 		    /* fall through */
 		case ss_accepting:
 		    set_error(sock,-ECONNRESET); /* ERESTARTSYS ? */
-		    send_release_complete(sock->call_ref,0);
+		    send_release_complete(sock->sig,sock->call_ref,0);
 		    SEND_ERROR(sock->id,sock->error);
 		    STOP_TIMER(sock);
 		    free_sock(sock);
@@ -421,12 +430,12 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 		    STOP_TIMER(sock);
 		    free_sock(sock);
 		    return;
-#ifdef Q2963_1
+#if defined(Q2963_1) || defined(DYNAMIC_UNI)
 		case ss_mod_req:
 #endif
 		    STOP_TIMER(sock);
 		    /* fall through */
-#ifdef Q2963_1
+#if defined(Q2963_1) || defined(DYNAMIC_UNI)
 		case ss_mod_lcl:
 		case ss_mod_rcv:
 		case ss_mod_fin_ok:
@@ -436,7 +445,7 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 		case ss_connected:
 		    diag(COMPONENT,DIAG_INFO,"Passive close (CR 0x%06X)",
 		      sock->call_ref);
-#ifdef Q2963_1
+#if defined(Q2963_1) || defined(DYNAMIC_UNI)
 		    if (timer_handler(sock->conn_timer) == on_T361)
 			STOP_TIMER(sock);
 #endif
@@ -446,14 +455,14 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 		case ss_indicated:
 		    /* fall through */
 		case ss_proceeding:
-		    send_release_complete(sock->call_ref,0);
+		    send_release_complete(sock->sig,sock->call_ref,0);
 		    new_state(sock,ss_zombie);
 		    /* fall through */
 		case ss_rel_ind:
 		    return;
 		default:
-		    send_release_complete(sock->call_ref,0); /* @@@ should
-			be ATM_CV_INCOMP_MSG */
+		    send_release_complete(sock->sig,sock->call_ref,0);
+			/* @@@ should be ATM_CV_INCOMP_MSG */
 		    break;
 	    }
 	    break;
@@ -477,7 +486,7 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 		    case ss_connecting:
 			set_error(sock,cause == ATM_CV_UNALLOC ?
 			  -EADDRNOTAVAIL : cause == ATM_CV_RES_UNAVAIL ||
-#if defined(UNI31) || defined(UNI40)
+#if defined(UNI31) || defined(UNI40) || defined(DYNAMIC_UNI)
 			  cause == ATM_CV_UCR_UNAVAIL_NEW ||
 #endif
 			  cause == ATM_CV_NO_ROUTE_DEST ? -EHOSTUNREACH :
@@ -498,12 +507,12 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 			STOP_TIMER(sock);
 			free_sock(sock);
 			return;
-#ifdef Q2963_1
+#if defined(Q2963_1) || defined(DYNAMIC_UNI)
 		    case ss_mod_req:
 #endif
 			STOP_TIMER(sock);
 			/* fall through */
-#ifdef Q2963_1
+#if defined(Q2963_1) || defined(DYNAMIC_UNI)
 		    case ss_mod_lcl:
 		    case ss_mod_rcv:
 		    case ss_mod_fin_ok:
@@ -513,7 +522,7 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 		    case ss_connected:
 			diag(COMPONENT,DIAG_INFO,"Passive close (CR 0x%06X)",
 			  sock->call_ref);
-#ifdef Q2963_1
+#if defined(Q2963_1) || defined(DYNAMIC_UNI)
 			if (timer_handler(sock->conn_timer) == on_T361)
 			    STOP_TIMER(sock);
 #endif
@@ -542,7 +551,7 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 
 		vpci = q_fetch(&in_dsc,QF_vpi);
 		sock->pvc.sap_family = AF_ATMPVC;
-		sock->pvc.sap_addr.itf = get_itf(&vpci);
+		sock->pvc.sap_addr.itf = get_itf(sock->sig,&vpci);
 		sock->pvc.sap_addr.vpi = vpci;
 		sock->pvc.sap_addr.vci = q_fetch(&in_dsc,QF_vci);
 		diag(COMPONENT,DIAG_DEBUG,"ITF.VPI.VCI: %d.%d.%d",
@@ -553,11 +562,12 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 	case ATM_MSG_NOTIFY:
 		/* silently ignore this junk */
 	    return;
-#ifdef Q2963_1
+#if defined(Q2963_1) || defined(DYNAMIC_UNI)
 /*
  * Buglet ahead: should actually test "call_state"
  */
 	case ATM_MSG_MODIFY_REQ:
+	    if (!(sock->sig->uni & S_Q2963_1)) goto _default;
 	    if (sock->state != ss_connected || sock->owner) break;
 	    sock->new_qos = sock->qos;
 	    if (q_present(&in_dsc,QF_fw_pcr_01))
@@ -569,6 +579,7 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 	    new_state(sock,ss_mod_rcv);
 	    return;
 	case ATM_MSG_MODIFY_ACK:
+	    if (!(sock->sig->uni & S_Q2963_1)) goto _default;
 	    if (sock->state != ss_mod_req) break;
 	    STOP_TIMER(sock);
 	    sock->qos = sock->new_qos;
@@ -578,6 +589,7 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 	    new_state(sock,ss_mod_fin_ok);
 	    return;
 	case ATM_MSG_MODIFY_REJ:
+	    if (!(sock->sig->uni & S_Q2963_1)) goto _default;
 	    if (sock->state != ss_mod_req) break;
 	    STOP_TIMER(sock);
 	    sock->error = -EAGAIN;
@@ -586,60 +598,63 @@ static void uni_call(SOCKET *sock,unsigned char mid)
 	    new_state(sock,ss_mod_fin_fail);
 	    return;
 	case ATM_MSG_CONN_AVAIL:
+	    if (!(sock->sig->uni & S_Q2963_1)) goto _default;
 	    if (sock->state != ss_connected || !sock->owner) break;
 	    STOP_TIMER(sock);
 	    send_kernel(sock->id,0,as_modify,ATM_MF_SET,NULL,NULL,NULL,NULL,
 	      &sock->qos);
 	    new_state(sock,ss_mod_fin_ack);
 	    return;
+	_default: /* jump here if we don't want to understand a message */
 #endif
 	default:
 	    diag(COMPONENT,DIAG_WARN,"Bad signaling message %d",mid);
-	    send_status(sock,0,ATM_CV_UNKNOWN_MSG_TYPE,mid);
+	    send_status(sock->sig,sock,0,ATM_CV_UNKNOWN_MSG_TYPE,mid);
 	    return;
     }
     diag(COMPONENT,DIAG_WARN,
       "Signaling message %s is incompatible with state %s/%s (%d?%d)",
       mid2name(mid),state_name[sock->state],cs_name[sock->call_state],
       (int) sock->state,(int) sock->call_state);
-    send_status(sock,0,ATM_CV_INCOMP_MSG,mid);
+    send_status(sock->sig,sock,0,ATM_CV_INCOMP_MSG,mid);
 }
 
 
-void clear_all_calls(void)
+void clear_all_calls(SIG_ENTITY *sig)
 {
     SOCKET *curr,*next;
 
     for (curr = sockets; curr; curr = next) {
 	next = curr->next;
-	if (curr->call_state != cs_null) uni_call(curr,ATM_MSG_RESTART);
+	if (curr->sig == sig && curr->call_state != cs_null)
+	    uni_call(curr,ATM_MSG_RESTART);
     }
 }
 
 
-void clear_all_calls_on_T309(void)
+void clear_all_calls_on_T309(SIG_ENTITY *sig)
 {
-    clear_all_calls();
+    clear_all_calls(sig);
     t309 = NULL;
 }
 
 
-void saal_failure(void)
+void saal_failure(SIG_ENTITY *sig)
 {
     SOCKET *curr,*next;
 
     trace_msg("SAAL went down");
     for (curr = sockets; curr; curr = next) {
 	next = curr->next;
-	if (curr->call_state != cs_null)
+	if (curr->sig == sig && curr->call_state != cs_null)
 		if (curr->call_state != cs_active)
 		    uni_call(curr,ATM_MSG_RESTART);
-		else if (!t309) t309 = start_timer(T309_TIME,on_T309,NULL);
+		else if (!t309) t309 = start_timer(T309_TIME,on_T309,sig);
     }
 }
 
 
-void saal_okay(void)
+void saal_okay(SIG_ENTITY *sig)
 {
     SOCKET *curr;
 
@@ -657,23 +672,24 @@ void saal_okay(void)
 	Q_DSC dsc;
 	int size;
 
-	clear_all_calls();
+	clear_all_calls(sig);
 	q_create(&dsc,q_buffer,MAX_Q_MSG);
 	q_assign(&dsc,QF_msg_type,QMSG_RESTART);
 	q_assign(&dsc,QF_call_ref,0);
 	q_assign(&dsc,QF_rst_class,ATM_RST_ALL_VC);
-	if ((size = q_close(&dsc)) >= 0) to_signaling(q_buffer,size);
+	if ((size = q_close(&dsc)) >= 0) to_signaling(sig,q_buffer,size);
     }
 #endif
     if (!t309) return;
     stop_timer(t309);
     t309 = NULL;
     for (curr = sockets; curr; curr = curr->next)
-	if (curr->call_state != cs_null) send_status_enq(curr);
+	if (curr->sig == sig && curr->call_state != cs_null)
+	    send_status_enq(curr);
 }
 
 
-static void process_uni(void *msg)
+static void process_uni(SIG_ENTITY *sig,void *msg)
 {
     SOCKET *curr;
     unsigned long call_ref;
@@ -692,69 +708,70 @@ static void process_uni(void *msg)
 		    int vpi,vci;
 
 		    if (!q_present(&in_dsc,QG_conn_id)) {
-			send_status(NULL,0L,ATM_CV_MAND_IE_MISSING,
+			send_status(sig,NULL,0L,ATM_CV_MAND_IE_MISSING,
 			  ATM_IE_CONN_ID);
 			return;
 		    }
 		    vpi = q_fetch(&in_dsc,QF_vpi);
 		    vci = q_fetch(&in_dsc,QF_vci);
 		    for (curr = sockets; curr; curr = curr->next)
-			if (curr->pvc.sap_addr.vpi == vpi &&
+			if (curr->sig == sig && curr->pvc.sap_addr.vpi == vpi &&
 			  curr->pvc.sap_addr.vci == vci) break;
 		    if (!curr) {
-			send_status(NULL,0L,ATM_CV_INVALID_IE,
+			send_status(sig,NULL,0L,ATM_CV_INVALID_IE,
 			  ATM_IE_CONN_ID);
 			return;
 		    }
 		    uni_call(curr,mid);
-		    send_restart_ack(vpi,vci);
+		    send_restart_ack(sig,vpi,vci);
 		}
 		break;
 	    case ATM_RST_ALL_VC:
-		clear_all_calls();
-		send_restart_ack(0,0);
+		clear_all_calls(sig);
+		send_restart_ack(sig,0,0);
 		break;
 	    default:
-		send_status(NULL,0L,ATM_CV_INVALID_IE,ATM_IE_RESTART);
+		send_status(sig,NULL,0L,ATM_CV_INVALID_IE,ATM_IE_RESTART);
 	}
 	return;
     }
     if (!(call_ref & 0x7fffff)) {
-	return; /* bad things happen ... @@@ */
+	diag(COMPONENT,DIAG_ERROR,"unrecognized global call ref");
+	return;
     }
     for (curr = sockets; curr; curr = curr->next)
-	if (curr->call_ref == call_ref) break;
-    diag(COMPONENT,DIAG_DEBUG,"FROM SAAL: %s (0x%02X) CR 0x%06lx for 0x%lx",
-      mid2name(((unsigned char *) msg)[5]),((unsigned char *)msg)[5],call_ref,
-      curr ? curr->id : 0);
+	if (curr->sig == sig && curr->call_ref == call_ref) break;
+    diag(COMPONENT,DIAG_DEBUG,"FROM SAAL %d.%d.%d: %s (0x%02X) CR 0x%06lx for "
+      "0x%lx",S_PVC(sig),mid2name(((unsigned char *) msg)[5]),
+      ((unsigned char *)msg)[5],call_ref,curr ? curr->id : 0);
     if (mid == ATM_MSG_SETUP) {
-	if (!curr) setup_call(call_ref);
+	if (!curr) setup_call(sig,call_ref);
 	return;
     }
     if (mid == ATM_MSG_STATUS_ENQ) {
-	send_status(curr,call_ref,ATM_CV_RESP_STAT_ENQ);
+	send_status(curr->sig,curr,call_ref,ATM_CV_RESP_STAT_ENQ);
 	return;
     }
     if (curr && q_present(&in_dsc,QF_ep_ref) && mid != ATM_MSG_ADD_PARTY &&
       mid != ATM_MSG_DROP_PARTY_ACK)
 	if (curr->ep_ref != q_fetch(&in_dsc,QF_ep_ref)) {
-	    send_drop_party_ack(call_ref,q_fetch(&in_dsc,QF_ep_ref),
+	    send_drop_party_ack(sig,call_ref,q_fetch(&in_dsc,QF_ep_ref),
 	      ATM_CV_INV_EPR);
 	    return;
 	}
     if (!curr || curr->call_state == cs_null) {
 	if (mid != ATM_MSG_REL_COMP)
 	    if (mid != ATM_MSG_STATUS)
-		send_release_complete(call_ref,ATM_CV_INV_CR);
+		send_release_complete(sig,call_ref,ATM_CV_INV_CR);
 	    else if (q_fetch(&in_dsc,QF_call_state) != (int) cs_null)
-		    send_release_complete(call_ref,ATM_CV_INCOMP_MSG);
+		    send_release_complete(sig,call_ref,ATM_CV_INCOMP_MSG);
 	return;
     }
     uni_call(curr,mid);
 }
 
 
-static void abort_call(unsigned char *msg,int size)
+static void abort_call(SIG_ENTITY *sig,unsigned char *msg,int size)
 {
     SOCKET *curr;
     unsigned long call_ref;
@@ -768,21 +785,21 @@ static void abort_call(unsigned char *msg,int size)
     diag(COMPONENT,DIAG_ERROR,"can't parse message - aborting the call "
       "(CR 0x%06lx)",call_ref);
     for (curr = sockets; curr; curr = curr->next)
-        if (curr->call_ref == call_ref) {
+        if (curr->sig == sig && curr->call_ref == call_ref) {
 	    uni_call(curr,ATM_MSG_RESTART);
 	    break;
 	}
-    send_release_complete(call_ref,ATM_CV_PROTOCOL_ERROR);
+    send_release_complete(sig,call_ref,ATM_CV_PROTOCOL_ERROR);
 }
 
 
-void to_uni(void *msg,int size)
+void to_uni(SIG_ENTITY *sig,void *msg,int size)
 {
     if (q_open(&in_dsc,msg,size) < 0) {
-	abort_call(msg,size);
+	abort_call(sig,msg,size);
 	return;
     }
-    process_uni(msg);
+    process_uni(sig,msg);
     if (q_close(&in_dsc) < 0)
 	diag(COMPONENT,DIAG_ERROR,"q_close returned <0 in to_uni");
 }
